# To do:
# Hammering prevention
#  - Peer finding
#  - Request trading
# Negative cache
# Max-age cache
# Remove dateutil/req from pagegetter
# Remove pytz/req from pagegetter

from twisted.python.failure import Failure

from uuid import uuid4
from twisted.internet import task
from twisted.internet import reactor
from twisted.internet.defer import Deferred, DeferredList

from pagegetter import RequestQueuer, PageGetter
from unicodeconverter import convertToUTF8

from aws import AmazonS3, AmazonSDB

from twisted.web import server
from twisted.web.resource import Resource
import time

import inspect

from twisted.internet import reactor

from twisted.web import static

import re
import os

from aws import sdb_now, sdb_now_add

from resources.template import TemplateResource
from resources.exposed import ExposedResource
from resources.data import DataResource
from resources.control import ControlResource
from resources.peer import PeerResource

from twisted.internet import threads

import cPickle
import sys


import pprint
PrettyPrinter = pprint.PrettyPrinter(indent=4)

import logging
import logging.handlers

from timeoffset import getTimeOffset
from networkaddress import getNetworkAddress

logger = logging.getLogger("main")

def evaluateBoolean( b ):
    if b == "false" or b == "False":
        return False
    elif b == "true" or b == "True":
        return True
    elif b == "no" or b == "No":
        return False
    elif b == "yes" or b == "Yes":
        return True
    else:
        try:
            return bool( int(b) )
        except:
            return False

class AWSpider:

    """
    Amazon S3, Amazon EC2 based web spider.
    """
    
    paused = False
    plugins = []
    start_time = time.time()
    
    public_ip = None
    local_ip = None
    network_information = {}
    
    uuid = uuid4().hex
    reserved_arguments = ["reservation_function_name", "reservation_created", "reservation_next_request", "reservation_error"]
    exposed_function_resources = {}
    functions = {}
    exposed_functions = []
    peers = {}
    peer_uuids = []
    
    def __init__( self, 
                aws_access_key_id, 
                aws_secret_access_key, 
                aws_s3_cache_bucket, 
                aws_sdb_reservation_domain, 
                aws_s3_storage_bucket=None,
                aws_sdb_coordination_domain=None, 
                port=8080, 
                max_requests=50, 
                log_directory=None, 
                log_level="debug",
                web_admin=True,
                web_admin_port=None,
                time_offset=None,
                peer_check_interval=5,
                reservation_check_interval=5):
        
        self.time_offset = time_offset
        self.peer_check_interval = int(peer_check_interval)
        self.reservation_check_interval = int(reservation_check_interval)
        
        self.port = int(port)
        self.network_information["port"] = self.port
        
        self.rq = RequestQueuer( max_requests=int(max_requests) )
        
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_s3_cache_bucket = aws_s3_cache_bucket
        self.aws_s3_storage_bucket = aws_s3_storage_bucket
        self.aws_sdb_reservation_domain = aws_sdb_reservation_domain
        self.aws_sdb_coordination_domain = aws_sdb_coordination_domain
        
        self.resource = Resource()
        
        if web_admin_port is None:
            web_admin_port = self.port
        else:
            web_admin_port = int(web_admin_port)

        if web_admin:
            self.resource.putChild('', TemplateResource("index" ) )
            self.resource.putChild('static', static.File( os.path.join(os.path.dirname(__file__), 'web_static') ) )
            self.resource.putChild('control', ControlResource( self ) )
          
        self.resource.putChild('data', DataResource( self ) )
        self.resource.putChild('peer', PeerResource( self ) )
        self.function_resource = Resource()
        self.resource.putChild( "function", self.function_resource )
        self.site = server.Site( self.resource )
        reactor.listenTCP(self.port, self.site)

        if web_admin and web_admin_port != self.port:
            self.web_admin_site = server.Site( self.resource )
            reactor.listenTCP( web_admin_port, self.web_admin_site )

        self.s3 = AmazonS3( self.aws_access_key_id, self.aws_secret_access_key, self.rq )
        self.sdb = AmazonSDB( self.aws_access_key_id, self.aws_secret_access_key, self.rq )
        
        self.pg = PageGetter( self.s3, self.aws_s3_cache_bucket, rq=self.rq )

        if log_directory is None:
            handler = logging.StreamHandler()
        else:
            handler = logging.handlers.TimedRotatingFileHandler(os.path.join(log_directory, 'spider.log'), when='D', interval=1)
            
        formatter = logging.Formatter("%(levelname)s: %(message)s %(pathname)s:%(lineno)d")
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        
        log_level = log_level.lower()
        log_levels = {
            "debug":logging.DEBUG, 
            "info":logging.INFO, 
            "warning":logging.WARNING, 
            "error":logging.ERROR,
            "critical":logging.CRITICAL
        }
        if log_level in log_levels:
            logger.setLevel(log_levels[log_level])
        else:
            logger.setLevel(logging.DEBUG)
    
        logger.info("Successfully loaded the Spider's configuration.")

    def showReservation( self, uuid ):
        d = self.sdb.getAttributes( self.aws_sdb_reservation_domain, uuid )
        return d

    def getServerData( self ):    
        
        running_time = time.time() - self.start_time
        cost = (self.sdb.box_usage * .14) * (60*60*24*30.4) / (running_time)
        
        data = {
            "paused":self.paused,
            "running_time":running_time,
            "cost":cost
        }
        
        logger.debug("Got server data:\n%s" % PrettyPrinter.pformat(data) )

        return data

    def getPage( self, *args, **kwargs ):
        return self.pg.getPage( *args, **kwargs )
    
    def getExposedFunctionDetails( self ):
        functions = []
        for key in self.exposed_functions:
            function = [key, {}]
            function[1]["interval"] = self.functions[key]["interval"]
            function[1]["required_arguments"] = self.functions[key]["required_arguments"]
            function[1]["optional_arguments"] = self.functions[key]["optional_arguments"]
            functions.append( function )
        
        functions.sort(lambda x,y:cmp(x[0], y[0]))
        
        logger.info("Got exposed function details:\n%s" % PrettyPrinter.pformat(functions) )
        
        return functions

    
    def createReservation( self, function_name, **kwargs ):
        
        if not isinstance( function_name, str ):
            for key in self.functions:
                if self.functions[key]["function"] == function_name:
                    function_name = key
                    break            
        
        if function_name not in self.functions:
            raise Exception("Function %s does not exist." % function_name )
    
        function = self.functions[ function_name ]
    
        filtered_kwargs = {}
        for key in function["required_arguments"]:
            if key in kwargs:
                filtered_kwargs[key] = convertToUTF8( kwargs[key] )
            else:
                raise Exception("Required parameter '%s' not found. Required parameters are %s. Optional parameters are %s." % (key, function["required_arguments"], function["optional_arguments"] ))

        for key in function["optional_arguments"]:
            if key in kwargs:
                filtered_kwargs[key] = convertToUTF8( kwargs[key] )
        
        if function["interval"] > 0:
            reserved_arguments = {}
            reserved_arguments["reservation_function_name"] = function_name
            reserved_arguments["reservation_created"] = sdb_now(offset=self.time_offset)
            reserved_arguments["reservation_next_request"] = reserved_arguments["reservation_created"]
            reserved_arguments["reservation_error"] = "0"
            
            all_arguments = {}
            all_arguments.update( reserved_arguments )
            all_arguments.update( filtered_kwargs )
            
            uuid = uuid4().hex
            logger.debug( "Creating reservation on SimpleDB for %s, %s." % (function_name, uuid))

            a = self.sdb.putAttributes( self.aws_sdb_reservation_domain, uuid, all_arguments )
            a.addCallback( self._createReservationCallback, function_name, uuid )
            a.addErrback( self._createReservationErrback, function_name, uuid )
                        
            if "call_immediately" in kwargs and not evaluateBoolean( kwargs["call_immediately"] ):    
                d = DeferredList([a], consumeErrors=True)
            else:
                logger.debug( "Calling %s in a thread immediately with arguments:\n%s" % (function_name, PrettyPrinter.pformat(filtered_kwargs) ) )
                b = Deferred()
                b.addCallback(self._callExposedFunctionWrapper, function["function"], filtered_kwargs, function_name, uuid)
                reactor.callInThread( b.callback, None )
                d = DeferredList([a,b], consumeErrors=True)
            
            d.addCallback( self._createReservationCallback2, function_name, uuid )
            d.addErrback( self._createReservationErrback2, function_name, uuid )
            return d
            
        else:
            logger.debug( "Calling %s in a thread immediately with arguments:\n%s" % (function_name, PrettyPrinter.pformat(filtered_kwargs) ) )
            d = Deferred()
            d.addCallback(self.callExposedFunctionImmediately, function["function"], filtered_kwargs, function_name)
            reactor.callInThread( d.callback, None )
            return d

    def _callExposedFunctionWrapper( self, data, func, kwargs, function_name, uuid ):
        return self.callExposedFunction( func, kwargs, function_name, uuid )

    def _createReservationCallback( self, data, function_name, uuid ):
        logger.debug( "Created reservation on SimpleDB for %s, %s." % (function_name, uuid))
        return uuid
    
    def _createReservationErrback( self, error, function_name, uuid ):
        logger.error( "Unable to create reservation on SimpleDB for %s:%s, %s.\n" % (function_name, uuid, error) )
        return error
    
    def _createReservationCallback2( self, data, function_name, uuid ):
        
        for row in data:
            if row[0] == False:
                raise row[1]
                
        if len(data) == 1:
            return {data[0][1]:{}}
        else:
            return {data[0][1]:data[1][1]}
            
    def _createReservationErrback2( self, error, function_name, uuid ):
        logger.error( "Unable to create reservation for %s:%s, %s.\n" % (function_name, uuid, error) )
        return error
    
    def callExposedFunctionImmediately( self, data, func, kwargs, function_name ):

        try:
            d = func( **kwargs )
        except Exception, e:
            return self._callExposedFunctionImmediatelyErrback( Failure(exc_value=sys.exc_value, exc_type=sys.exc_type, exc_tb=sys.exc_traceback), function_name )
            
        if isinstance( d, Deferred ):            
            d.addCallback( self._callExposedFunctionImmediatelyCallback, function_name )
            d.addErrback( self._callExposedFunctionImmediatelyErrback, function_name )
            return d
        else:
            return self._callExposedFunctionImmediatelyCallback( d, function_name )
        
    def _callExposedFunctionImmediatelyCallback( self, data, function_name ):
        logger.info("Function %s returned successfully:%s" % (function_name, data) )
        return data
        
    def _callExposedFunctionImmediatelyErrback( self, error, function_name ):
        logger.error( "Error with immediate callback of %s.\n%s\n%s" % (function_name, error, error.value.__dict__) )

        return error
            
    def expose( self, *args, **kwargs ):
        return self.makeCallable( expose=True, *args, **kwargs )
        
    def makeCallable( self, func, interval=0, name=None, expose=False ):
            
        argspec = inspect.getargspec( func )
        arguments = argspec[0]
        if len(arguments) > 0 and arguments[0:1][0] == 'self':
            arguments.pop(0)
        
        kwarg_defaults = argspec[3]
        if kwarg_defaults is None:
            kwarg_defaults = []
        
        required_arguments = arguments[ 0:len(arguments) - len(kwarg_defaults) ]
        optional_arguments = arguments[ len(arguments) - len(kwarg_defaults):]

        if hasattr(func, "im_class") and name is None:
            function_name = "%s/%s" % ( func.im_class.__name__, func.__name__)
        elif name is not None:
            function_name = name
        else:
            function_name = func.__name__
        
        function_name = function_name.lower()
        
        for key in required_arguments:
            if key in self.reserved_arguments:
                message = "Required argument name '%s' used in function %s is reserved." % (key, function_name)
                logger.error( message )
                raise Exception( message )
                
        for key in optional_arguments:
            if key in self.reserved_arguments:
                message = "Optional argument name '%s' used in function %s is reserved." % (key, function_name)
                logger.error( message )
                raise Exception( message )
        
        self.functions[ function_name ] = {
            "function":func,
            "interval":interval,
            "required_arguments":required_arguments,
            "optional_arguments":optional_arguments
        }
        
        logger.info( "Function %s is now callable by the Spider." % function_name )
        
        if expose:
            
            self.exposed_functions.append( function_name )
            
            er = ExposedResource( self, function_name )
        
            function_name_parts = function_name.split("/")
            if len( function_name_parts ) > 1:
                if function_name_parts[0] in self.exposed_function_resources:
                    r = self.exposed_function_resources[ function_name_parts[0] ]
                else:
                    r = Resource()
                    self.exposed_function_resources[ function_name_parts[0] ] = r
                self.function_resource.putChild(function_name_parts[0], r )
                r.putChild( function_name_parts[1], er)
            else:
                self.function_resource.putChild(function_name_parts[0], er )
            
            logger.info( "Function %s is now available via the Spider's web interface." % function_name )
            
        return function_name
        
    def start( self ):
        reactor.callWhenRunning( self._start )
        
    def _start( self ):
        logger.critical( "Checking S3 and SimpleDB setup, getting time offset for sync." )
        
        deferreds = []
        
        # Make sure the specified S3 cache bucket works.
        
        deferreds.append( self._checkAndCreateS3Bucket( self.aws_s3_cache_bucket ) )

        if self.aws_s3_storage_bucket is not None:
            deferreds.append( self._checkAndCreateS3Bucket( self.aws_s3_storage_bucket ) )

        deferreds.append( self._checkAndCreateSDBDomain( self.aws_sdb_reservation_domain ) )    

        if self.aws_sdb_coordination_domain is not None:
            deferreds.append( self._checkAndCreateSDBDomain( self.aws_sdb_coordination_domain ) )
            deferreds.append( self.getNetworkAddress() )
        
        if self.time_offset is None:
            deferreds.append( self.getTimeOffset() )
        
        d = DeferredList(deferreds, consumeErrors=True )
        d.addCallback( self._startCallback )

    def _startCallback( self, data ):   
        reactor.addSystemEventTrigger('before', 'shutdown', self._cleanupBeforeReactorShutdown)     
        logger.critical( "Starting Spider UUID: %s" % self.uuid)
        self.queryloop = task.LoopingCall( self.query )
        self.queryloop.start( self.reservation_check_interval )
        if self.aws_sdb_coordination_domain is not None:
            self.coordinateloop = task.LoopingCall( self.coordinate )
            self.coordinateloop.start( self.peer_check_interval )  
            reactor.callLater( 10, self.peerCheckRequest )          
        #self.paused = True
        pass
   
    def getNetworkAddress( self ):
        d = getNetworkAddress()
        d.addCallback( self._getNetworkAddressCallback )
        d.addErrback( self._getNetworkAddressErrback )
        return d
                
    def _getNetworkAddressCallback( self, data  ):   
        if "public_ip" in data:
            self.public_ip = data["public_ip"]
            self.network_information["public_ip"] = self.public_ip
            
        if "local_ip" in data:
            self.local_ip = data["local_ip"]  
            self.network_information["local_ip"] = self.local_ip
         
    def _getNetworkAddressErrback( self, error ):
        logger.critical( "Could not get network address." )
        self.shutdown()
            
    def getTimeOffset( self ):
        d = getTimeOffset()
        d.addCallback( self._getTimeOffsetCallback )
        d.addErrback( self._getTimeOffsetErrback )
        return d
        
    def _getTimeOffsetCallback( self, time_offset ):
        self.time_offset = time_offset
        logger.info( "Got time offset for sync: %s" % self.time_offset )
    
    def _getTimeOffsetErrback( self, error ):
        if self.time_offset is None:
            logger.critical( "Could not get time offset for sync." )
            self.shutdown()
    
    def _checkAndCreateS3Bucket( self, bucket_name ):
        d = self.s3.getBucket( bucket_name )
        d.addErrback( self._checkAndCreateS3BucketErrback, bucket_name ) 
        return d
        
    def _checkAndCreateS3BucketErrback( self, error, bucket_name ):
        if int(error.value.status) == 404:
            logger.info( "Bucket %s does not exist, creating." % bucket_name )     
            d = self.s3.putBucket( bucket_name )
            d.addErrback( self._checkAndCreateS3BucketErrback2, bucket_name )
            return d
            
        logger.critical( "Could not find or create S3 bucket '%s'." % bucket_name )
        self.shutdown()
        
    def _checkAndCreateS3BucketErrback2( self, error, bucket_name ):
        logger.critical( "Could not create S3 bucket '%s'" % bucket_name )
        self.shutdown()

    def _checkAndCreateSDBDomain( self, domain_name ):    
        d = self.sdb.domainMetadata( domain_name )
        d.addErrback( self._checkAndCreateSDBDomainErrback, domain_name )     
        return d

    def _checkAndCreateSDBDomainErrback( self, error, domain_name ):
        
        if int(error.value.status) == 400:  
            logger.info( "Domain %s does not exist, creating." % domain_name )
            d = self.sdb.createDomain( domain_name  )
            d.addErrback( self._checkAndCreateSDBDomainErrback2, domain_name )
            return d
        
        logger.critical( "Could not find or create SDB domain '%s'." % domain_name )
        self.shutdown()
        
    def _checkAndCreateSDBDomainErrback2( self, error, domain_name ):
        logger.critical( "Could not create SDB domain '%s'" % domain_name )
        self.shutdown()
    
    def pause(self):
        logger.critical( "Pausing Spider." )
        self.paused = True
        self.queryloop.stop() 
        self.coordinateloop.stop()
        
    def resume(self):
        logger.critical( "Resuming Spider." )
        self.paused = False
        self.queryloop.start( self.reservation_check_interval ) 
        if self.aws_sdb_coordination_domain is not None:
            self.coordinateloop.start( self.peer_check_interval)
    
    def setReservationError( self, uuid, function_name="Unknown" ):
        logger.error( "Setting error on reservation %s, %s." % (function_name, uuid) )
        d = self.sdb.putAttributes( self.aws_sdb_reservation_domain, uuid, {"reservation_error":"1"}, replace=["reservation_error"] )
        d.addCallback( self._setReservationErrorCallback, uuid )
        d.addErrback( self._setReservationErrorErrback, uuid )
        return d
        
    def _setReservationErrorCallback( self, data, function_name, uuid ):
        logger.error( "Error set on reservation %s, %s." % (function_name, uuid ) )
        
    def _setReservationErrorErrback( self, error, function_name, uuid ):
        logger.error( "Unable to set error on reservation %s, %s.\n%s" % (function_name, uuid, error) )
    
    def deleteFunctionReservations( self, function_name ):
        logger.info( "Deleting reservations for function %s." % (function_name) )
        d = self.sdb.select( "SELECT itemName() FROM %s WHERE reservation_function_name='%s'" % ( self.aws_sdb_reservation_domain, function_name) )
        d.addCallback( self._deleteFunctionReservationsCallback, function_name )
        d.addErrback( self._deleteFunctionReservationsErrback, function_name )
        return d
    
    def _deleteFunctionReservationsCallback( self, data, function_name ):
        logger.debug( "Found reservations for function %s: %s" % (function_name, data) )
        deferreds = []
        for uuid in data:
            deferreds.append( self.sdb.delete(self.aws_sdb_reservation_domain, uuid) )
        d = DeferredList(deferreds, consumeErrors = True )
        d.addCallback( self._deleteFunctionReservationsCallback2, function_name )
        return d
    
    def _deleteFunctionReservationsCallback2( self, data, function_name ):
        try:
            for row in data:
                if row[0] == False:
                    raise row[1] 
        except:
            return self._deleteFunctionReservationsErrback( Failure(exc_value=sys.exc_value, exc_type=sys.exc_type, exc_tb=sys.exc_traceback), function_name )
        
        logger.debug( "Deleted reservations for function %s." % (function_name) )
        
    def _deleteFunctionReservationsErrback( self, error, function_name ):
        logger.error( "Error deleting reservations for %s.\n%s" ) % (function_name, error)
        
    def deleteReservation( self, uuid, function_name="Unknown" ):
        logger.info( "Deleting reservation %s, %s." % (function_name, uuid) )
        d = self.sdb.delete( self.aws_sdb_reservation_domain, uuid )
        d.addCallback( self._deleteReservationCallback, function_name, uuid )
        d.addErrback( self._deleteReservationErrback, function_name, uuid )
        return d
        
    def _deleteReservationCallback( self, data, function_name, uuid ):
        logger.info( "Reservation %s, %s successfully deleted." % (function_name, uuid) )
        
    def _deleteReservationErrback( self, error, function_name, uuid ):
        logger.error( "Error deleting reservation %s, %s.\n%s" ) % (function_name, uuid, error)

    def query( self ):
        
        if len(self.functions) > 0:
            reservation_function_names = []
            for function_name in self.functions.keys():
                 reservation_function_names.append( "reservation_function_name='%s'" % function_name )
            reservation_function_names =  "INTERSECTION " + " OR ".join( reservation_function_names )
        else:
            reservation_function_names = ""
        
        sql = "SELECT * FROM %s WHERE reservation_next_request < '%s' INTERSECTION reservation_error != '1' %s" % ( self.aws_sdb_reservation_domain, sdb_now(offset=self.time_offset), reservation_function_names )
        
        #logger.debug( "Querying SimpleDB, \"%s\"" % sql )
        
        d = self.sdb.select( sql )
        d.addCallback( self._queryCallback )
        d.addErrback( self._queryErrback )
    
    def _queryErrback( self, error ):
        logger.error( "Unable to query SimpleDB.\n%s" % error )
        
    def _queryCallback( self, data ):
        
        for uuid in data:
            
            kwargs_raw = {}
            reserved_arguments = {}
            
            for key in data[ uuid ]:
                if key in self.reserved_arguments:
                    reserved_arguments[ key ] = data[ uuid ][key][0]
                else:
                    kwargs_raw[key] = data[ uuid ][key][0]
            
            if "reservation_function_name" not in reserved_arguments:
                logger.error( "Reservation %s does not have a function name." % uuid )
                self.deleteReservation( uuid )
                continue

            function_name = reserved_arguments["reservation_function_name"]

            if "reservation_created" not in reserved_arguments:
                logger.error( "Reservation %s, %s does not have a created time." % (function_name, uuid) )
                self.deleteReservation( uuid, function_name=function_name )
                continue            
            
            if "reservation_next_request" not in reserved_arguments:
                logger.error( "Reservation %s, %s does not have a next request time." % (function_name, uuid) )
                self.deleteReservation( uuid, function_name=function_name )
                continue                

            if "reservation_error" not in reserved_arguments:
                logger.error( "Reservation %s, %s does not have an error flag." % (function_name, uuid) )
                self.deleteReservation( uuid, function_name=function_name )
                continue
            
            if reserved_arguments["reservation_function_name"] in self.functions:
                exposed_function = self.functions[ function_name  ]
            else:
                logger.error( "Could not find resource %s." % function_name )
                continue
            
            d = self.sdb.putAttributes( self.aws_sdb_reservation_domain, uuid, {"reservation_next_request":sdb_now_add(exposed_function["interval"], offset=self.time_offset) }, replace=["reservation_next_request"] )
            d.addCallback( self._setNextRequestCallback, function_name, uuid)
            d.addErrback( self._setNextRequestErrback, function_name, uuid )
            
            kwargs = {}
            for key in kwargs_raw:
                if key in exposed_function["required_arguments"]:
                    kwargs[ key ] = kwargs_raw[ key ]
                if key in exposed_function["optional_arguments"]:
                    kwargs[ key ] = kwargs_raw[ key ]
            
            has_reqiured_arguments = True
            for key in exposed_function["required_arguments"]:
                if key not in kwargs:
                    has_reqiured_arguments = False
                    logger.error( "%s, %s does not have required argument %s." % (function_name, uuid, key) ) 
                    
            if not has_reqiured_arguments:
                self.setReservationError( uuid )
                continue
            
            reactor.callInThread(self.callExposedFunction, exposed_function["function"], kwargs, function_name, uuid  )

    def callExposedFunction( self, func, kwargs, function_name, uuid ):
        
        logger.debug( "Calling %s, %s." % (function_name, uuid) )
        
        try:
            d = func( **kwargs )
        except:         
            return self._exposedFunctionErrback( Failure(exc_value=sys.exc_value, exc_type=sys.exc_type, exc_tb=sys.exc_traceback), function_name, uuid )
        
        if isinstance(d, Deferred ):
            d.addCallback( self._exposedFunctionCallback, function_name, uuid )
            d.addErrback( self._exposedFunctionErrback, function_name, uuid )
            return d
        else:
            return self._exposedFunctionCallback( d, function_name, uuid )
            
    def _setNextRequestCallback(self, data, function_name, uuid ):
        logger.debug( "Set next request for %s, %s on on SimpleDB." % (function_name, uuid) )
        
    def _setNextRequestErrback(self, error, function_name, uuid ):
        logger.error( "Unable to set next request for %s, %s on SimpleDB.\n%s" % (function_name, uuid, error.value) )
    
    def get( self, uuid ):
        logger.debug( "Getting %s from S3." % uuid )
        d = self.s3.getObject( self.aws_s3_storage_bucket, uuid )
        d.addCallback( self._getCallback, uuid )
        d.addErrback( self._getErrback, uuid )
        return d
        
    def _getCallback( self, data, uuid ):
        logger.debug( "Got %s from S3." % (uuid) ) 
        return cPickle.loads( data["response"] )
    
    def _getErrback( self, error, uuid ):
        logger.error( "Could not get %s from S3.\n%s" % (uuid, error) ) 
        return error
    
    def _exposedFunctionCallback( self, data, function_name, uuid ):
        
        if self.aws_s3_storage_bucket is not None:
            if data is None:
                logger.debug( "Recieved None for %s, %s." % (function_name, uuid) )
                return None
            else:    
                logger.debug( "Putting result for %s, %s on S3." % (function_name, uuid) )
                pickled_data = cPickle.dumps( data )
                d = self.s3.putObject( self.aws_s3_storage_bucket, uuid, pickled_data, contentType="text/plain", gzip=True )
                d.addErrback( self._exposedFunctionErrback2, function_name, uuid )
                d.addCallback( self._exposedFunctionCallback2, data )
                return d
    
    def _exposedFunctionCallback2( self, s3_callback_data, data ):
        return data
    
    def _exposedFunctionErrback( self, error, function_name, uuid ):
        logger.error( "Error with %s, %s.\n%s" % (function_name, uuid, error) )
    
    def _exposedFunctionErrback2( self, error, function_name, uuid ):
        logger.error( "Could not put results of %s, %s on S3.\n%s" % (function_name, uuid, error) )

    
    def _cleanupBeforeReactorShutdown( self ):
        logger.debug( "Cleaning up before reactor shutdown." )
        
        deferreds = []
        
        if self.aws_sdb_coordination_domain is not None:
            d = self.sdb.delete( self.aws_sdb_coordination_domain, self.uuid )
            d.addCallback( self._cleanupBeforeReactorShutdownCallback )
            return d
    
    def _cleanupBeforeReactorShutdownCallback( self, data  ):
        logger.debug( "Deleted uuid from %s" % self.aws_sdb_coordination_domain )
        return self.peerCheckRequest()
    
    def peerCheckRequest( self ):

        logger.debug( "Signaling peers." )

        deferreds = []

        for uuid in self.peers:
            if uuid != self.uuid:
                logger.debug( "Signaling %s to check peers." % self.peers[uuid]["uri"] )
                d = self.getPage( self.peers[uuid]["uri"] + "/peer/check", cache= -1 )
                d.addCallback( self._peerCheckRequestCallback, self.peers[uuid]["uri"] )
                deferreds.append( d )

        if len(deferreds) > 0:
            return DeferredList(deferreds)

    def _peerCheckRequestCallback( self, data, uri ):
        logger.debug( "Got %s/peer/check." % uri )
    
    def shutdown( self ):
        logger.critical( "Stopping server." )
        reactor.callLater( 1, reactor.stop )
    
    def coordinate( self ):
        #logger.debug( "Peers: %s" % self.peers )
        attributes = { "created":sdb_now(offset=self.time_offset) }
        attributes.update( self.network_information )
        self.sdb.putAttributes( self.aws_sdb_coordination_domain, self.uuid, attributes, replace=attributes.keys() )
        reactor.callLater( 5, self.checkPeers )
    
    def checkPeers( self ):
        sql = "SELECT public_ip, local_ip, port FROM %s WHERE created > '%s'" % ( self.aws_sdb_coordination_domain, sdb_now_add( self.peer_check_interval * -2, offset=self.time_offset) )
        #logger.debug( "Querying SimpleDB, \"%s\"" % sql )
        d = self.sdb.select( sql )
        d.addCallback( self._checkPeersCallback )
        d.addErrback( self._checkPeersErrback )
        
    def _checkPeersCallback( self, discovered ):

        existing_peers = set( self.peers.keys() )
        discovered_peers = set( discovered.keys() )
        new_peers = discovered_peers - existing_peers
        old_peers = existing_peers - discovered_peers
        
        deferreds = []
        
        for uuid in new_peers:
            if uuid == self.uuid:
                self.peers[uuid] = {
                    "uri":"http://127.0.0.1:%s" % self.port,
                    "local_ip":"127.0.0.1",
                    "port":self.port,
                    "active":True
                }
            else:
                deferreds.append( self.verifyPeer( uuid, discovered[uuid] ) )
        
        for uuid in old_peers:
            logger.debug( "Removing peer %s" % uuid )
            
            if uuid in self.peers:
                del self.peers[ uuid ]
        
        if len( new_peers ) > 0:
            if len(deferreds) > 0:
                d = DeferredList(deferreds, consumeErrors=True)
                d.addCallback( self._checkPeersCallback2 )
            else:
                self._checkPeersCallback2( None ) #Just found ourself.
        elif len( old_peers ) > 0:
            self._checkPeersCallback2( None )
        else:
            pass # No old, no new.
        
    def _checkPeersCallback2( self, data ):
        
        logger.debug( "Re-organizing peers." )
        
        for uuid in self.peers:
            if "local_ip" in self.peers[uuid]:
                self.peers[uuid]["uri"] = "http://%s:%s" % (self.peers[uuid]["local_ip"], self.peers[uuid]["port"] )
                self.peers[uuid]["active"] = True
            elif "public_ip" in self.peers[uuid]:
                self.peers[uuid]["uri"] = "http://%s:%s" % (self.peers[uuid]["public_ip"], self.peers[uuid]["port"] )
                self.peers[uuid]["active"] = True
            else:
                logger.error("Peer %s has no local or public IP. This should not happen." % uuid )
        
        self.peer_uuids = self.peers.keys()
        self.peer_uuids.sort()
        
#        print self.peers 
        logger.debug( "Peers updated to: %s" % self.peers )
        # Go through the peers and reorganize the group.
        pass
        
    def _checkPeersErrback( self, error ):
        logger.error( "Could not query SimpleDB for peers: %s" % str(error) )

    def verifyPeer( self, uuid, peer ):
        
        logger.debug( "Verifying peer %s" % uuid )
        
        deferreds = []
        
        if "port" in peer:
            port = int(peer["port"][0])
        else:
            port = self.port

        if uuid not in self.peers:
            self.peers[ uuid ] = {}
            
        self.peers[ uuid ]["active"] = False
        self.peers[ uuid ]["port"] = port
        
        if "local_ip" in peer:
            local_ip = peer["local_ip"][0]
            local_url = "http://%s:%s/data/server" % (local_ip, port)
            d = self.getPage( local_url, timeout=5, cache=-1 )
            d.addCallback( self._verifyPeerLocalIPCallback, uuid, local_ip, port )
            deferreds.append( d )
            
        if "public_ip" in peer:
            public_ip = peer["public_ip"][0]
            public_url = "http://%s:%s/data/server" % (public_ip, port)
            d = self.getPage( public_url, timeout=5, cache=-1  )         
            d.addCallback( self._verifyPeerPublicIPCallback, uuid, public_ip, port )
            deferreds.append( d )
        
        if len(deferreds) > 0:
            d = DeferredList( deferreds, consumeErrors=True )
            return d
        else:
            return None
    
    def _verifyPeerLocalIPCallback( self, data, uuid, local_ip, port ):
        logger.debug( "Verified local IP for %s" % uuid )
        self.peers[ uuid ]["local_ip"] = local_ip
        
    def _verifyPeerPublicIPCallback( self, data, uuid, public_ip, port ):
        logger.debug( "Verified public IP for %s" % uuid )
        self.peers[ uuid ]["public_ip"] = public_ip
        




from twisted.internet import reactor
from twisted.internet import defer
from twisted.web.client import HTTPClientFactory, _parse
from unicodeconverter import convertToUTF8, convertToUnicode
import cPickle
import sha
import time
import dateutil.parser
import datetime
import pytz
import urllib

import logging
logger = logging.getLogger("main")

class RequestQueuer():

    def __init__(self, max_requests=50):
    
        self.max_requests = max_requests
            
        self.pending = []
        self.active = 0

    def checkActive(self):
        
        while self.active < self.max_requests:
            
            if len( self.pending ) > 0:
                
                req = self.pending.pop(0)
                
                self.active = self.active + 1

                d = self._getPage(  
                    req["url"], 
                    method=req["method"], 
                    postdata=req["postdata"],
                    headers=req["headers"], 
                    agent=req["agent"], 
                    timeout=req["timeout"], 
                    cookies=req["cookies"],
                    followRedirect=req["followRedirect"]
                )

                d.addCallback( self.requestComplete, req["deferred"] ).addErrback( self.requestError, req["deferred"] )

            else:
                break

    def requestComplete( self, response, deferred ):
        self.active = self.active - 1
        self.checkActive()
        deferred.callback( response )
        return None
    
    def requestError( self, error, deferred ):
        self.active = self.active - 1
        self.checkActive()
        deferred.errback( error )  
        return None
    
    def getPage(self, url, last_modified=None, etag=None, method='GET', postdata=None, headers=None, agent="Offbeat Guides", timeout=60, cookies=None, followRedirect=1, prioritize=False ):

        if headers is None:
            headers = {}

        if last_modified is not None:
            headers['If-Modified-Since'] = time.strftime( "%a, %d %b %Y %T %z", dateutil.parser.parse(last_modified).timetuple() )

        if etag is not None:
            headers["If-None-Match"] = etag

        deferred = defer.Deferred()
                
        url = convertToUTF8( url )
        
        req = { 
            "url":url, 
            "method":method, 
            "postdata":postdata,
            "headers":headers, 
            "agent":agent, 
            "timeout":timeout, 
            "cookies":cookies,
            "followRedirect":followRedirect,
            "deferred":deferred 
        }
                
        if prioritize:
            self.pending.insert( 0, req )
        else:
            self.pending.append( req )
            
        self.checkActive()

        return deferred

    def _getPage(self, url, method='GET', postdata=None, headers=None, agent="AWSpider", timeout=60, cookies=None, followRedirect=1 ):

        if headers is None:
            headers = {}

        scheme, host, port, path = _parse(url)

        factory = HTTPClientFactory(
            url, 
            method=method, 
            postdata=postdata,
            headers=headers, 
            agent=agent, 
            timeout=timeout, 
            cookies=cookies,
            followRedirect=followRedirect
        )

        if scheme == 'https':
            from twisted.internet import ssl
            contextFactory = ssl.ClientContextFactory()
            reactor.connectSSL(host, port, factory, contextFactory, timeout=timeout)
        else:
            reactor.connectTCP(host, port, factory, timeout=timeout)

        factory.deferred.addCallback( self._getPageComplete, factory )
        factory.deferred.addErrback( self._getPageError, factory )
        return factory.deferred

    def _getPageComplete( self, response, factory ):
        return {"response":response, "headers":factory.response_headers, "status":int(factory.status), "message":factory.message }

    def _getPageError( self, error, factory ):
        
        if "response_headers" in factory.__dict__ and factory.response_headers is not None:
            error.value.headers = factory.response_headers
        
        return error

class PageGetter:
    
    def __init__( self, s3, aws_s3_bucket, rq=None ):
        
        self.s3 = s3
        self.aws_s3_bucket = aws_s3_bucket
        
        if rq is None:
            self.rq = RequestQueuer()
        else:
            self.rq = rq
    
    def getPage(self, url, method='GET', postdata=None, headers=None, agent="AWSpider", timeout=60, cookies=None, followRedirect=1, hash_url=None, cache=0, metadata=None, prioritize=True ):

        cache=int(cache)
        
        if cache not in [-1,0,1]:
            raise Exception("Unknown caching mode.")

        url = convertToUTF8( url )
        if hash_url is not None:
            hash_url = convertToUTF8( hash_url )

        if metadata is None:
            metadata = {}
        
        if isinstance( postdata, dict ):
            for key in postdata:
                postdata[key] = convertToUTF8( postdata[key] )
            postdata = urllib.urlencode( postdata )
        
        if method.lower() == "post":
            if headers is None:
                headers = {"content-type":"application/x-www-form-urlencoded"}
            else:
                headers["content-type"] = "application/x-www-form-urlencoded"
                
        if method.lower() != "get":
            d = self.rq.getPage( url, method=method, postdata=postdata, headers=headers, agent=agent, timeout=timeout, cookies=cookies, followRedirect=followRedirect, prioritize=prioritize )
            return d       
        
        parameters_to_be_hashed = cPickle.dumps( [headers, agent, cookies] )
        
        if hash_url is None:
            metadata["url"] = url
            request_hash = sha.new( url + parameters_to_be_hashed ).hexdigest()
        else:
            metadata["url"] = hash_url
            request_hash = sha.new( hash_url + parameters_to_be_hashed ).hexdigest()
        
        if cache == -1:
            logger.debug( "Getting request %s for URL %s." % (request_hash, url) )
            d = self.rq.getPage( url, method="GET", postdata=postdata, headers=headers, agent=agent, timeout=timeout, cookies=cookies, followRedirect=followRedirect, prioritize=prioritize )
            d.addCallback( self._storeData, request_hash )
            return d
            
        if cache == 0:
            logger.debug( "Checking S3 Head object request %s for URL %s." % (request_hash, url) )
            d = self.s3.headObject( self.aws_s3_bucket, request_hash )
            d.addCallback( self._s3HeadObjectCallback, url, request_hash, postdata, headers, agent, timeout, cookies, followRedirect, prioritize )
            d.addErrback( self._s3HeadObjectErrback, url, request_hash, postdata, headers, agent, timeout, cookies, followRedirect, prioritize )        
            return d
        
        if cache == 1:
            logger.debug( "Getting S3 object request %s for URL %s." % (request_hash, url) )
            d = self.s3.getObject( self.aws_s3_bucket, request_hash )
            d.addCallback( self._s3GetObjectCallback, request_hash )
            d.addErrback( self._s3HeadObjectErrback, url, request_hash, postdata, headers, agent, timeout, cookies, followRedirect, prioritize )       
            return d             
        
    def _s3HeadObjectCallback( self, data, url, request_hash, postdata, headers, agent, timeout, cookies, followRedirect, prioritize ):
        
        logger.debug( "Got S3 Head object request %s for URL %s." % (request_hash, url) )
    
        try:
            if "cache-expires" in data["headers"]:
                expires = dateutil.parser.parse(data["headers"]["cache-expires"][0])
                now = datetime.datetime.now( pytz.timezone("UTC") )
                if expires < now:
                    logger.debug( "Stored data request %s for URL %s is not stale. Getting from S3." % (request_hash, url) )
                    d = self.s3.getObject( self.aws_s3_bucket, request_hash )
                    d.addCallback( self._s3GetObjectCallback, request_hash )
                    d.addErrback( self._s3GetObjectErrback, None, request_hash )
                    return d                
        except Exception, e:
            logger.error( str(e) )
            
        request_keywords = {
            "method":"GET", 
            "postdata":postdata, 
            "headers":headers, 
            "agent":agent, 
            "timeout":timeout, 
            "cookies":cookies, 
            "followRedirect":followRedirect, 
            "prioritize":prioritize
        }
        
        if "cache-etag" in data["headers"]:
            request_keywords["etag"] = data["headers"]["cache-etag"][0]
        
        if "cache-last-modified" in data["headers"]:
            request_keywords["last_modified"] = data["headers"]["cache-last-modified"][0]
        
        #print "Requesting page with etag, last_modified headers"
        logger.debug( "Requesting %s for URL %s etag and last-modified headers." % (request_hash, url) )
        d = self.rq.getPage( url, **request_keywords )
        d.addCallback( self._s3HeadObjectCallback2, request_hash, url )
        d.addErrback( self._s3HeadObjectErrback2, request_hash, url )
        return d
        
    def _s3HeadObjectCallback2( self, data, request_hash, url ):
        logger.debug( "Got request %s for URL %s." % (request_hash, url) )
        return data
    
    def _s3HeadObjectErrback2( self, error, request_hash, url ):
        
        if error.value.status == "304":
            #print "Page hasn't been modified"
            logger.debug( "Request %s for URL %s hasn't been modified since it was last downloaded. Getting data from S3." % (request_hash, url) )
            d = self.s3.getObject( self.aws_s3_bucket, request_hash )
            d.addCallback( self._s3GetObjectCallback, request_hash )
            d.addErrback( self._s3GetObjectErrback, error, request_hash )
            return d
        else:
            logger.error( "Unable to get request %s for URL %s.\n%s" % (request_hash, url, error) )
            return error
    
    def _s3GetObjectCallback( self, data, request_hash ):
        
        logger.debug( "Got request %s from S3." % (request_hash) )
        
        #print "Got S3 object"
        data["status"] = 304
        data["message"] = "Not Modified"
        if "cache-expires" in data["headers"]:
            data["headers"]["expires"] = data["headers"]["cache-expires"]
            del data["headers"]["cache-expires"]
        
        if "cache-etag" in data["headers"]:
            data["headers"]["etag"] = data["headers"]["cache-etag"]
            del data["headers"]["cache-etag"]

        if "cache-last-modified" in data["headers"]:
            data["headers"]["last-modified"] = data["headers"]["cache-last-modified"]
            del data["headers"]["cache-last-modified"]
        
        return data
    
    def _s3GetObjectErrback( self, error, page_error, request_hash ):
        
        logger.error( "Error getting data from S3 for request %s.\n%s\n%s" % (request_hash, page_error, error) )
        
        if page_error is None:
            return error
        else:
            return page_error
            
    def _s3HeadObjectErrback(self, error, url, request_hash, postdata, headers, agent, timeout, cookies, followRedirect, prioritize):
        # No header, let's get the object!
        
        logger.debug( "Unable to find header for request %s on S3, fetching from %s." % (request_hash, url) )
        
        d = self.rq.getPage( url, method="GET", postdata=postdata, headers=headers, agent=agent, timeout=timeout, cookies=cookies, followRedirect=followRedirect, prioritize=prioritize )
        d.addCallback( self._storeData, request_hash )
        return d
        
    def _storeData( self, data, request_hash ):
        
        logger.debug("Writing data for request %s to S3." % request_hash )

        headers = {}
        
        if "cache-control" in data["headers"]: 
            if "no-cache" in data["headers"]["cache-control"][0]:
                return data
        
        if "expires" in data["headers"]:
            headers["cache-expires"] = data["headers"]["expires"][0]

        if "etag" in data["headers"]:
            headers["cache-etag"] = data["headers"]["etag"][0]
            
        if "last-modified" in data["headers"]:
            headers["cache-last-modified"] = data["headers"]["last-modified"][0]

        if "content-type" in data["headers"]:
            contentType = data["headers"]["content-type"][0]
        
        self.s3.putObject( self.aws_s3_bucket, request_hash, data["response"], contentType=contentType, headers=headers )
        return data
        
        



from twisted.internet import reactor
from twisted.internet.protocol import ClientCreator
from twisted.conch.telnet import Telnet
from twisted.internet.defer import Deferred

import datetime
import time
import calendar
import random

import logging
logger = logging.getLogger("main")

import pytz

timeservers=["time.nist.gov", "time-a.nist.gov", "time-b.nist.gov", "time-nw.nist.gov", "nist1-ny.WiTime.net", "nist1-dc.WiTime.net", "nist1.aol-va.symmetricom.com", "nist1.columbiacountyga.gov", "nist.expertsmi.com", "nist.netservicesgroup.com", "time-a.timefreq.bldrdoc.gov", "time-c.timefreq.bldrdoc.gov", "utcnist.colorado.edu", "utcnist2.colorado.edu", "ntp-nist.ldsbc.edu", "nist1.aol-ca.symmetricom.com", "nist1.symmetricom.com", "nist1-sj.WiTime.net", "nist1-la.WiTime.net"]

class SimpleTelnet(Telnet):
    
    def __init__( self, *args, **kwargs ):
        self.deferred = Deferred()
        self.data = []
        Telnet.__init__(self, *args, **kwargs )
        
    def dataReceived(self, data):
        self.data.append( data )
    
    def connectionLost( self, reason ):
        self.deferred.callback( "".join(self.data) )

def getTimeOffset():
    
    client = ClientCreator(reactor, SimpleTelnet)
    server = timeservers.pop(0)
    logger.debug( "Requesting time from %s." % server )
    d = client.connectTCP(server, 13, timeout=5)
    d.addCallback( _getTimeOffsetCallback, server )
    d.addErrback( _getTimeOffsetErrback, 0 )
    return d

def _getTimeOffsetErrback( error, count ):
    if count < 5:
        client = ClientCreator(reactor, SimpleTelnet)
        server = timeservers.pop()
        logger.debug( "Attempt %s failed, requesting time from %s." % (count + 1, server) )
        d = client.connectTCP(server, 13, timeout=5)
        d.addCallback( _getTimeOffsetCallback, server )
        d.addErrback( _getTimeOffsetErrback, count + 1 ) 
        return d   
    else:
        logger.debug( "Could not fetch time after %s attempts." % count )
        return error
        
def _getTimeOffsetCallback( simple_telnet, server ):
    logger.debug( "Connected to time server %s." % server )
    simple_telnet.deferred.addCallback( _getTimeOffsetCallback2, server )
    return simple_telnet.deferred
    
def _getTimeOffsetCallback2( data, server ):

    logger.debug( "Got time from %s." % server )

    t = datetime.datetime( 
        2000 + int(data[7:9]), 
        int(data[10:12]), 
        int(data[13:15]),
        int(data[16:18]),
        int(data[19:21]),
        int(data[22:24]) )
        
    offset = calendar.timegm( t.timetuple() ) - time.time()
    return offset

def getTimeOffsetTest():
    d = getTimeOffset()
    d.addCallback(_getTimeOffsetTestCallback)
    
def _getTimeOffsetTestCallback( offset ):
    print offset
    print time.time()
    print time.time() + offset
    reactor.stop()
    
if __name__ == "__main__":
    import logging.handlers
    handler = logging.StreamHandler()
    formatter = logging.Formatter("%(levelname)s: %(message)s %(pathname)s:%(lineno)d")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    reactor.callWhenRunning( getTimeOffsetTest )     
    reactor.run()


from twisted.web.resource import Resource
from twisted.web import server
import traceback
import simplejson
from twisted.python.failure import Failure
from twisted.internet import reactor

class PeerResource(Resource):
    
    isLeaf = True
    
    def __init__( self, spider ):
        
        self.spider = spider
        Resource.__init__(self)
    
    def render(self, request):
        
        request.setHeader('Content-type', 'text/javascript; charset=UTF-8')
        
        if len(request.postpath) > 0:
            if request.postpath[0] == "check":
                reactor.callLater( 5, self.spider.checkPeers )
                return simplejson.dumps( True )         
                    
    def _successResponse( self, data ):
        return simplejson.dumps( True )

    def _errorResponse( self, error ):

        reason = str(error.value)
        tb = traceback.format_exc( traceback.extract_tb(error.tb) )

        return simplejson.dumps( {"error":reason, "traceback":tb} )

    def _immediateResponse(self, data, request ):           
        request.write( data )
        request.finish()


from twisted.python.failure import Failure
from twisted.web.resource import Resource
from awspider.aws import sdb_now
from uuid import uuid4
import simplejson
from twisted.web import server
import traceback
import sys
import cStringIO, gzip

class ExposedResource(Resource):
    
    isLeaf = True
    
    def __init__( self, spider, function_name ):
        self.spider = spider
        self.function_name = function_name
        Resource.__init__(self)
    
    def render(self, request):
        request.setHeader('Content-type', 'text/javascript; charset=UTF-8')
        
        kwargs = {}
        for key in request.args:
            kwargs[key] = request.args[key][0]
            
        try:
            d = self.spider.createReservation( self.function_name, **kwargs )
        except:
            return self._errorResponse(Failure(exc_value=sys.exc_value, exc_type=sys.exc_type, exc_tb=sys.exc_traceback))
            
        d.addCallback( self._successResponse )
        d.addErrback( self._errorResponse )
        d.addCallback( self._immediateResponse, request )
        return server.NOT_DONE_YET

    def _successResponse( self, data ):
        return simplejson.dumps(data)
        
    def _errorResponse( self, error ):

        reason = str(error.value)
        tb = traceback.format_exc( traceback.extract_tb(error.tb) )

        return simplejson.dumps( {"error":reason, "traceback":tb} )
        
    def _immediateResponse( self, data, request ):    
        
        encoding = request.getHeader("accept-encoding")
        
        if encoding and "gzip" in encoding:

            zbuf = cStringIO.StringIO()
            zfile = gzip.GzipFile(None, 'wb', 9, zbuf)
            if isinstance( data, unicode ):
                zfile.write( unicode(data).encode("utf-8") )
            elif isinstance( data, str ):
                zfile.write( unicode(data, 'utf-8' ).encode("utf-8") )
            else:
                zfile.write( unicode(data).encode("utf-8") )

            zfile.close()
            request.setHeader("Content-encoding","gzip")
            request.write( zbuf.getvalue() )
        else:
            request.write( data )
        request.finish()

    



from twisted.web.resource import Resource
from twisted.web import server
import traceback
import simplejson
from twisted.python.failure import Failure
class ControlResource(Resource):
    
    isLeaf = True
    
    def __init__( self, spider ):
        
        self.spider = spider
        Resource.__init__(self)
    
    def render(self, request):
        
        request.setHeader('Content-type', 'text/javascript; charset=UTF-8')
        
        if len(request.postpath) > 0:
            if request.postpath[0] == "query":
                self.spider.query()
                return simplejson.dumps( True )
            elif request.postpath[0] == "shutdown":
                self.spider.shutdown()
                return simplejson.dumps( True )
            elif request.postpath[0] == "pause":
                self.spider.pause()
                return simplejson.dumps( True )
            elif request.postpath[0] == "resume":
                self.spider.resume()    
                return simplejson.dumps( True )
         
                    
    def _successResponse( self, data ):
        return simplejson.dumps( True )

    def _errorResponse( self, error ):

        reason = str(error.value)
        tb = traceback.format_exc( traceback.extract_tb(error.tb) )

        return simplejson.dumps( {"error":reason, "traceback":tb} )

    def _immediateResponse(self, data, request ):           
        request.write( data )
        request.finish()


from twisted.web.resource import Resource
import simplejson
from twisted.internet.defer import DeferredList
from twisted.web import server
import traceback
from datetime import datetime
from twisted.python.failure import Failure
import cStringIO, gzip



class DataResource(Resource):
    
    isLeaf = True
    
    def __init__( self, spider ):
        
        self.spider = spider
        Resource.__init__(self)
    
    def render(self, request):
        
        request.setHeader('Content-type', 'text/javascript; charset=UTF-8')
        
        if len(request.postpath) > 0:
            if request.postpath[0] == "server":
                return simplejson.dumps( self.spider.getServerData() )
            elif request.postpath[0] == "exposed_function_details":
                return simplejson.dumps( self.spider.getExposedFunctionDetails() )
            elif request.postpath[0] == "get":
                if "uuid" in request.args:
                    get_deferreds = []
                    for uuid in request.args["uuid"]:
                        get_deferreds.append( self.spider.get( uuid ) )
                    d = DeferredList( get_deferreds, consumeErrors = True )
                    d.addCallback( self._getCallback, request.args["uuid"] )
                    d.addCallback( self._successResponse )
                    d.addErrback( self._errorResponse )
                    d.addCallback( self._immediateResponse, request )
                    
                    return server.NOT_DONE_YET
                else:
                    return self._errorResponse(Failure(exc_value=Exception("Parameter UUID is required.")))
            elif request.postpath[0] == "show_reservation":
                if "uuid" in request.args:
                    d = self.spider.showReservation( request.args["uuid"][0] )
                    d.addCallback( self._successResponse )
                    d.addErrback( self._errorResponse )
                    d.addCallback( self._immediateResponse, request )
                    return server.NOT_DONE_YET
                else:
                    return self._errorResponse(Failure(exc_value=Exception("Parameter UUID is required.")))
            elif request.postpath[0] == "delete_reservation":
                if "uuid" in request.args:
                    d = self.spider.deleteReservation( request.args["uuid"][0] )
                    d.addCallback( self._successResponse )
                    d.addErrback( self._errorResponse )
                    d.addCallback( self._immediateResponse, request )
                    return server.NOT_DONE_YET
                else:
                    return self._errorResponse(Failure(exc_value=Exception("Parameter UUID is required.")))
            elif request.postpath[0] == "delete_function_reservations": 
                if "function_name" in request.args:
                    d = self.spider.deleteFunctionReservations( request.args["function_name"][0] )
                    d.addCallback( self._successResponse )
                    d.addErrback( self._errorResponse )
                    d.addCallback( self._immediateResponse, request )
                    return server.NOT_DONE_YET           
                else:
                    return self._errorResponse(Failure(exc_value=Exception("Parameter function_name is required."))) 
                                                            
    def _getCallback( self, data, uuids ):
        response = {}
        for i in range(0, len(uuids) ):
            if data[i][0] == True:
                response[ uuids[i] ] = data[i][1]
            else:
                response[ uuids[i] ] = {"error":str(data[i][1].value)}
                
        return response
        
    def _successResponse( self, data ):
        return simplejson.dumps(data)

    def _errorResponse( self, error ):
        reason = str(error.value)
        tb = traceback.format_exc( traceback.extract_tb(error.tb) )
        return simplejson.dumps( {"error":reason, "traceback":tb } )

    def _immediateResponse(self, data, request ):
        encoding = request.getHeader("accept-encoding")

        if encoding and "gzip" in encoding:

            zbuf = cStringIO.StringIO()
            zfile = gzip.GzipFile(None, 'wb', 9, zbuf)
            if isinstance( data, unicode ):
                zfile.write( unicode(data).encode("utf-8") )
            elif isinstance( data, str ):
                zfile.write( unicode(data, 'utf-8' ).encode("utf-8") )
            else:
                zfile.write( unicode(data).encode("utf-8") )

            zfile.close()
            request.setHeader("Content-encoding","gzip")
            request.write( zbuf.getvalue() )
        else:
            request.write( data )
        request.finish()


from ez_setup import use_setuptools
use_setuptools()

from setuptools import setup, find_packages
# Also requires python-dev and python-openssl
setup(

    name = "AWSpider",
    version = "0.1.5.10",

    packages = find_packages(),

    install_requires = ['twisted>=8.1', 'genshi>=0.5.1', 'python-dateutil>=1.4', 'pytz>=2009g', 'simplejson>=2.0.9', 'pyopenssl>=0.9'],
    include_package_data = True,
    
    # metadata for upload to PyPI
    author = "John Wehr",
    author_email = "johnwehr@gmail.com",
    description = "Amazon Web Services web crawler",
    license = "MIT License",
    keywords = "AWS amazon ec2 s3 simpledb twisted spider crawler",
    url = "http://github.com/wehriam/awspider/"
    
)



from unicodeconverter import convertToUTF8
from pagegetter import RequestQueuer
from time import strftime, gmtime
from datetime import datetime
import base64, hmac, sha, md5, urllib

import hashlib

import urllib

import xml.etree.cElementTree as ET

import time

import cStringIO
import gzip as gzip_package

import dateutil.parser

def sdb_now(offset=0):
    return str(int(offset + time.time())).zfill(11)

def sdb_now_add( i, offset=0 ):
    return str(int(offset + time.time() + i) ).zfill(11)

def sdb_parse_time(s, offset=0): # Parse a string into a SDB formatted timestamp
    return str(int(offset + time.mktime(dateutil.parser.parse( s ).timetuple()))).zfill(11)
    

    
def sdb_latitude( latitude ):
    adjusted = ( 90 + float( latitude ) ) * 100000
    return str( int( adjusted ) ).zfill(8)

def sdb_longitude( longitude ):
    adjusted = ( 180 + float( longitude) ) * 100000
    return str( int( adjusted ) ).zfill(8)

class AmazonS3:
    
    ACCEPTABLE_ERROR_CODES = [400, 403, 404, 409]
    
    def __init__(self, aws_access_key_id, aws_secret_access_key, rq=None ):
        
        # We're using a centralized Request Queuer when possible so as to not overwhelm the system.
        if rq is None:
            self.rq = RequestQueuer()
        else:
            self.rq = rq
        
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
    

    
    def getBucket( self, bucket ):
        bucket = convertToUTF8( bucket )
        host = "%s.s3.amazonaws.com" % bucket
        headers = {
            "Host": host
        }
        headers.update( self._getAuthorizationHeaders("GET", "", "", headers, "/" + bucket + "/") )
        url = "http://%s/" % host
        d = self.rq.getPage(url, method="GET", headers=headers )
        
        # Errback will try again a few times if it gets unusual (ie 503) errors.
        d.addErrback( self._genericErrback, url, method="GET", headers=headers )
        return d          

    def putBucket( self, bucket ):
        bucket = convertToUTF8( bucket )
        host = "%s.s3.amazonaws.com" % bucket
        headers = {
            "Host": host,
            "Content-Length": 0
        }
        headers.update( self._getAuthorizationHeaders("PUT", "", "", headers, "/" + bucket + "/") )
        url = "http://%s/" % host
        d = self.rq.getPage( url, method="PUT", headers=headers )
        
        # Errback will try again a few times if it gets unusual (ie 503) errors.
        d.addErrback( self._genericErrback, url, method="PUT", headers=headers )
        return d

    def deleteBucket( self, bucket ):
        bucket = convertToUTF8( bucket )
        host = "%s.s3.amazonaws.com" % bucket
        headers = {
            "Host": host,
            "Content-Length": 0
        }
        headers.update( self._getAuthorizationHeaders("DELETE", "", "", headers, "/" + bucket + "/") )
        url = "http://%s/" % host
        d = self.rq.getPage( url, method="DELETE", headers=headers )

        # Errback will try again a few times if it gets unusual (ie 503) errors.
        d.addErrback( self._genericErrback, url, method="DELETE", headers=headers )
        return d

    def headObject( self, bucket, key ):

        bucket = convertToUTF8( bucket )
        key = convertToUTF8( key )

        host = "%s.s3.amazonaws.com" % bucket
        headers = {
            "Host": host
        }
        headers.update( self._getAuthorizationHeaders("HEAD", "", "", headers, "/" + bucket + "/" + key ) )
        url = "http://%s/%s" % (host, key)
        d = self.rq.getPage(url, method="HEAD", headers=headers )

        # Errback will try again a few times if it gets unusual (ie 503) errors.
        d.addCallback( self._getObjectCallback )
        d.addErrback( self._genericErrback, url, method="HEAD", headers=headers )
        return d    

    def _headObjectCallback( self, data ):
        data["headers"] = dict( zip( map(lambda x:x.replace("x-amz-meta-",""), data["headers"].keys() ), data["headers"].values() ) )
        return data    

    def getObject( self, bucket, key ):
        
        bucket = convertToUTF8( bucket )
        key = convertToUTF8( key )
        
        host = "%s.s3.amazonaws.com" % bucket
        headers = {
            "Host": host
        }
        headers.update( self._getAuthorizationHeaders("GET", "", "", headers, "/" + bucket + "/" + key ) )
        url = "http://%s/%s" % (host, key)
        d = self.rq.getPage(url, method="GET", headers=headers )

        # Errback will try again a few times if it gets unusual (ie 503) errors.
        d.addCallback( self._getObjectCallback )
        d.addErrback( self._genericErrback, url, method="GET", headers=headers )
        return d    

    def _getObjectCallback( self, data ):
        
        if "content-encoding" in data["headers"]:
            if "gzip" in data["headers"]["content-encoding"]:
                compressedstream = cStringIO.StringIO( data["response"] ) 
                zfile = gzip_package.GzipFile( fileobj=compressedstream )
                data["response"] = zfile.read()
        
        data["headers"] = dict( zip( map(lambda x:x.replace("x-amz-meta-",""), data["headers"].keys() ), data["headers"].values() ) )
        return data
    
    def putObject( self, bucket, key, data, contentType=None, public=True, headers=None, gzip=False ):
        
        if contentType is None:
            contentType = "text/html"
        
        data = convertToUTF8( data ) 
        bucket = convertToUTF8( bucket )
        key = convertToUTF8( key )
        
        host = "%s.s3.amazonaws.com" % bucket
        
        if headers is None:
            headers = {}
        
        # Wrap any user-defined headers in with the Amazon custom header prefix.
        headers = dict( zip( map(lambda x:"x-amz-meta-" + x, headers.keys() ), headers.values() ) )
        
        if gzip:
            # Gzip that bastard!
            headers["content-encoding"] = "gzip"
            zbuf = cStringIO.StringIO()
            zfile = gzip_package.GzipFile(None, 'wb', 9, zbuf)
            zfile.write( data )
            zfile.close()
            data =  zbuf.getvalue()
        
        content_md5 = base64.encodestring( md5.new( data ).digest() ).strip()
        
        headers['x-amz-acl'] = 'public-read'
        
        #authorization_headers = self._getAuthorizationHeaders( 'PUT', content_md5, contentType, headers, "/" + bucket + "/" + key )
        generated_headers = { 'Content-Length': len( data ), 'Content-Type': contentType, 'Content-MD5': content_md5 }
        
        #headers.update( authorization_headers )
        headers.update( generated_headers )
        
        headers.update( self._getAuthorizationHeaders("PUT", content_md5, contentType, headers, "/" + bucket + "/" + key ) )
        
        url = "http://%s.s3.amazonaws.com/%s" % ( bucket, key )

        d = self.rq.getPage( url, method="PUT", headers=headers, postdata=data )

        d.addErrback( self._genericErrback, url, method="PUT", headers=headers, postdata=data )
        return d

    def deleteObject( self, bucket, key ):

        bucket = convertToUTF8( bucket )
        key = convertToUTF8( key )

        host = "%s.s3.amazonaws.com" % bucket
        headers = {
            "Host": host
        }
        headers.update( self._getAuthorizationHeaders("DELETE", "", "", headers, "/" + bucket + "/" + key ) )
        url = "http://%s/%s" % (host, key)
        d = self.rq.getPage(url, method="DELETE", headers=headers )

        # Errback will try again a few times if it gets unusual (ie 503) errors.
        d.addErrback( self._genericErrback, url, method="DELETE", headers=headers )
        return d
            
    def _genericErrback( self, error, url, method="GET", headers=None, postdata=None, count=0 ):
        
        if headers is None:
            headers = {}
        
        if "status" in error.value.__dict__:
            # 204, empty response but otherwise OK, as in the case of a delete.
            # Move on, nothing to see here.
            if int(error.value.status) == 204:
                return { "response":error.value.response, "status":int(error.value.status), "headers":error.value.response, "message":error.value.headers }

            # Something other than a 40x error. Something is wrong, but let's try
            # that again a few times.           
            elif int(error.value.status) not in self.ACCEPTABLE_ERROR_CODES and count < 3:

                d = self.rq.getPage( url, method=method, headers=headers, postdata=postdata )
                d.addErrback( self._genericErrback, url, method=method, headers=headers, postdata=postdata, count=count + 1 )
                return d

            # 404 or other normal error, pass it along.
            else:
                return error    
        else:
            return error
                
    def _canonalizeAmazonHeaders( self, headers ):
        keys = [k for k in headers.keys() if k.startswith("x-amz-")]
        keys.sort( key = str.lower)
        return "\n".join( map( lambda key: key + ":" + headers.get( key ), keys ) )

    def _getAuthorizationHeaders( self, method, contentHash, contentType, headers, resource ):
        date = strftime( u"%a, %d %b %Y %H:%M:%S +0000", gmtime() )
        amazon_headers = self._canonalizeAmazonHeaders( headers )
        if len( amazon_headers ):
            data = "%s\n%s\n%s\n%s\n%s\n%s" % ( method, contentHash, contentType, date, amazon_headers, resource )
        else:
            data = "%s\n%s\n%s\n%s\n%s%s" % ( method, contentHash, contentType, date, amazon_headers, resource )            
        authorization = "AWS %s:%s" % ( self.aws_access_key_id, base64.encodestring( hmac.new( self.aws_secret_access_key, data, sha ).digest() ).strip() )
        return { 'Authorization': authorization, 'Date': date }


 
 
class AmazonSDB:
    
    ACCEPTABLE_ERROR_CODES = []
    
    def __init__(self, aws_access_key_id, aws_secret_access_key, rq=None ):
        
        # We're using a centralized Request Queuer when possible so as to not overwhelm the system.
        if rq is None:
            self.rq = RequestQueuer()
        else:
            self.rq = rq
        
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.host = "sdb.amazonaws.com"
        self.box_usage = 0.0
        
    def request( self, parameters ):
        parameters = self._getAuthorization("GET", parameters )
        query_string = urllib.urlencode(parameters)
        url = "https://%s/?%s" % (self.host, query_string)
        d = self.rq.getPage(url, method="GET" )
        return d

    def etree_to_dict( self, etree ):
        children = etree.getchildren()
        if len(children) == 0:
            return etree.text
        children_dict = {}
        for element in children:
            tag = element.tag.replace("{http://sdb.amazonaws.com/doc/2007-11-07/}", "")
            if tag in children_dict:
                children_dict[ tag ].append( self.etree_to_dict( element ) )
            else:
                children_dict[ tag ] = [ self.etree_to_dict( element ) ]
        return children_dict

    def createDomain( self, domain ):
        parameters = {}
        parameters["Action"] = "CreateDomain"
        parameters["DomainName"] = domain
        d = self.request(parameters)
        d.addCallback( self._createDomainCallback )
        return d

    def _createDomainCallback( self, data ):
        
        xml = ET.fromstring(data["response"])
        self.box_usage += float( xml.find(".//{http://sdb.amazonaws.com/doc/2007-11-07/}BoxUsage").text )
        
        return True
    
    def deleteDomain( self, domain ):
        parameters = {}
        parameters["Action"] = "DeleteDomain"
        parameters["DomainName"] = domain
        d = self.request(parameters)
        d.addCallback( self._deleteDomainCallback )
        return d

    def _deleteDomainCallback( self, data ):
        
        xml = ET.fromstring(data["response"])
        self.box_usage += float( xml.find(".//{http://sdb.amazonaws.com/doc/2007-11-07/}BoxUsage").text )
        
        return True    

    def listDomains( self, next_token=None, previous_results=None ):
        parameters = {}
        parameters["Action"] = "ListDomains"
        if next_token is not None:
            parameters["NextToken"] = next_token
        d = self.request(parameters)
        d.addCallback( self._listDomainsCallback, previous_results=previous_results )
        return d

    def _listDomainsCallback( self, data, previous_results=None ):
        
        
        xml = ET.fromstring(data["response"])
        self.box_usage += float( xml.find(".//{http://sdb.amazonaws.com/doc/2007-11-07/}BoxUsage").text )
        
        xml_response = self.etree_to_dict( xml )

        if "DomainName" in xml_response["ListDomainsResult"][0]:
            results = xml_response["ListDomainsResult"][0]["DomainName"]
        else:
            results = []
        
        if previous_results is not None:
            results.extend( previous_results )
        
        if "NextToken" in xml_response["ListDomainsResult"]:
            return self.listDomains( next_token=xml_response["ListDomainsResult"][0]["NextToken"][0], previous_results=results )
        
        return results

    def domainMetadata( self, domain ):
        parameters = {}
        parameters["Action"] = "DomainMetadata"
        parameters["DomainName"] = domain
        d = self.request(parameters)
        d.addCallback( self._domainMetadataCallback )
        return d
        
    def _domainMetadataCallback( self, data):
        
        xml = ET.fromstring(data["response"])
        self.box_usage += float( xml.find(".//{http://sdb.amazonaws.com/doc/2007-11-07/}BoxUsage").text )
        
        xml_response = self.etree_to_dict( xml )
        return xml_response["DomainMetadataResult"][0]
    
    def putAttributes(self, domain, item_name, attributes, replace=None ):
        
        if replace is None:
            replace = []

        if not isinstance(replace, list):
            raise Exception("Replace parameter must be a list.")
        
        if not isinstance(attributes, dict):
            raise Exception("Attributes parameter must be a dictionary.")
        
        parameters = {}
        parameters["Action"] = "PutAttributes"
        parameters["DomainName"] = domain
        parameters["ItemName"] = item_name
        
        attributes_list = []
        for attribute in attributes.items():
            if isinstance(attribute[1], list):
                for value in attribute[1]:
                    attributes_list.append( (attribute[0], value) )
            else:
                attributes_list.append( attribute )
        
        i = 0
        for attribute in attributes_list:
            parameters["Attribute.%s.Name" % i] = attribute[0]
            parameters["Attribute.%s.Value" % i] = attribute[1]
            if attribute[0] in replace:
                parameters["Attribute.%s.Replace" % i] = "true"
            i += 1
        
        d = self.request(parameters)
        d.addCallback( self._putAttributesCallback )
        return d
        
    def _putAttributesCallback( self, data ):
        
        xml = ET.fromstring(data["response"])
        self.box_usage += float( xml.find(".//{http://sdb.amazonaws.com/doc/2007-11-07/}BoxUsage").text )
        
        return True
    
    def getAttributes( self, domain, item_name, attribute_name=None ):
        parameters = {}
        parameters["Action"] = "GetAttributes"
        parameters["DomainName"] = domain
        parameters["ItemName"] = item_name
        if attribute_name is not None:
            parameters["AttributeName"] = attribute_name
        d = self.request(parameters)
        d.addCallback( self._getAttributesCallback )
        return d
        
    def _getAttributesCallback( self, data ):
        print data
        xml = ET.fromstring(data["response"])
        self.box_usage += float( xml.find(".//{http://sdb.amazonaws.com/doc/2007-11-07/}BoxUsage").text )
        
        xml_response = self.etree_to_dict( xml )
        attributes = {}

        if xml_response["GetAttributesResult"][0] is None:
            raise Exception("Item does not exist.")
            
        for attribute in xml_response["GetAttributesResult"][0]['Attribute']:
            if attribute["Name"][0] not in attributes:
                attributes[ attribute["Name"][0] ] = []
            attributes[ attribute["Name"][0] ].extend( attribute["Value"] )
        return attributes

    def delete( self, domain, item_name ):
        return self.deleteAttributes(domain, item_name )

    def deleteAttributes(self, domain, item_name, attributes=None ):
        
        if attributes is None:
            attributes = {}
        
        if not isinstance(attributes, dict) and not isinstance(attributes, list):
            raise Exception("Attributes parameter must be a dictionary or a list.")

        parameters = {}
        parameters["Action"] = "DeleteAttributes"
        parameters["DomainName"] = domain
        parameters["ItemName"] = item_name
        
        if isinstance(attributes, dict):
            i = 0
            for key in attributes:
                parameters["Attribute.%s.Name" % i] = key
                parameters["Attribute.%s.Value" % i] = attributes[ key ]
                i += 1
                
        if isinstance(attributes, list):
            i = 0
            for key in attributes:
                parameters["Attribute.%s.Name" % i] = key
                i += 1
        
        d = self.request(parameters)
        d.addCallback( self._deleteAttributesCallback )
        return d

    def _deleteAttributesCallback( self, data ):
        
        xml = ET.fromstring(data["response"])
        self.box_usage += float( xml.find(".//{http://sdb.amazonaws.com/doc/2007-11-07/}BoxUsage").text )
        
        return True
    
    def select( self, select_expression, next_token=None, previous_results=None ):
        parameters = {}
        parameters["Action"] = "Select"
        parameters["SelectExpression"] = select_expression
        if next_token is not None:
            parameters["NextToken"] = next_token
        d = self.request(parameters)
        d.addCallback( self._selectCallback, select_expression=select_expression, previous_results=previous_results )
        return d

    def _selectCallback( self, data, select_expression=None, previous_results=None ):
        
        if previous_results is not None:
            results = previous_results
        else:
            results = {}
            
        xml = ET.fromstring(data["response"])
        self.box_usage += float( xml.find(".//{http://sdb.amazonaws.com/doc/2007-11-07/}BoxUsage").text )
                
        next_token_element = xml.find(".//{http://sdb.amazonaws.com/doc/2007-11-07/}NextToken")
        
        if next_token_element is not None:
            next_token = next_token_element.text
        else:
            next_token = None
            
        items = xml.findall(".//{http://sdb.amazonaws.com/doc/2007-11-07/}Item")
        results = {}
        for item in items:
            name = item.find("{http://sdb.amazonaws.com/doc/2007-11-07/}Name").text
            attributes = item.findall("{http://sdb.amazonaws.com/doc/2007-11-07/}Attribute")
            attribute_dict = {}
            for attribute in attributes:
                attribute_name = attribute.find("{http://sdb.amazonaws.com/doc/2007-11-07/}Name").text
                attribute_value = attribute.find("{http://sdb.amazonaws.com/doc/2007-11-07/}Value").text
                if attribute_name in attribute_dict:   
                    attribute_dict[ attribute_name ].append( attribute_value )
                else:
                    attribute_dict[ attribute_name ] = [ attribute_value ]
            results[ name ] = attribute_dict
        
        if next_token is not None:
            return self.select( select_expression, next_token=next_token, previous_results=results )

        return results        
    
    def safe_quote( self, s ):
        return urllib.quote( str(s), '-_.~' )
        
    def _canonicalize( self, parameters ):
        parameters = parameters.items()
        parameters.sort( lambda x,y:cmp(x[0],y[0]) )
        return "&".join( map( lambda x:"%s=%s" % ( self.safe_quote(x[0]), self.safe_quote(x[1]) ), parameters) )
        
    def _getAuthorization( self, method, parameters ):
        signature_parameters = { "AWSAccessKeyId":self.aws_access_key_id, "SignatureVersion":"2", "SignatureMethod":"HmacSHA256", 'Timestamp':datetime.utcnow().isoformat()[0:19]+"+00:00", "AWSAccessKeyId":self.aws_access_key_id, "Version":"2007-11-07" }
        signature_parameters.update( parameters )
        qs = self._canonicalize( signature_parameters )
        string_to_sign = "%(method)s\n%(host)s\n%(resource)s\n%(qs)s" % {
            "method":method,
            "host":self.host.lower(),
            "resource":"/",
            "qs":qs,
        }
          
        signature = base64.encodestring( hmac.new( self.aws_secret_access_key, string_to_sign, hashlib.sha256 ).digest() ).strip()
        signature_parameters.update( {'Signature': signature})
        return signature_parameters












